@@ -36,8 +36,8 @@
 
 #include "util.h"
 #include "sema.h"
-#include "io.h"
-
+#include "priv-io.h"
+#include "debug.h"
 
 /* We assume that a HANDLE can be represented by an int which should
    be true for all i386 systems (HANDLE is defined as void *) and
@@ -51,6 +51,7 @@
 
 #define READBUF_SIZE 4096
 #define WRITEBUF_SIZE 4096
+#define PIPEBUF_SIZE  4096
 #define MAX_READERS 20
 #define MAX_WRITERS 20
 
@@ -261,6 +262,12 @@ create_reader (HANDLE fd)
         free (c);
         return NULL;
     }    
+    else {
+      /* We set the priority of the thread higher because we know that
+         it only runs for a short time.  This greatly helps to increase
+         the performance of the I/O. */
+      SetThreadPriority (c->thread_hd, THREAD_PRIORITY_HIGHEST);
+    }
 
     return c;
 }
@@ -395,12 +402,11 @@ _gpgme_io_read ( int fd, void *buffer, size_t count )
     UNLOCK (c->mutex);
 
     DEBUG2 ("fd %d: got %d bytes\n", fd, nread );
+    if (nread > 0)
+      _gpgme_debug (2, "fd %d: got `%.*s'\n", fd, nread, buffer);
 
     return nread;
 }
-
-
-
 /*
  * The writer does use a simple buffering strategy so that we are
  * informed about write errors as soon as possible (i.e. with the the
@@ -514,6 +520,12 @@ create_writer (HANDLE fd)
         free (c);
         return NULL;
     }    
+    else {
+      /* We set the priority of the thread higher because we know that
+         it only runs for a short time.  This greatly helps to increase
+         the performance of the I/O. */
+      SetThreadPriority (c->thread_hd, THREAD_PRIORITY_HIGHEST);
+    }
 
     return c;
 }
@@ -600,6 +612,7 @@ _gpgme_io_write ( int fd, const void *buffer, size_t count )
     struct writer_context_s *c = find_writer (fd,1);
 
     DEBUG2 ("fd %d: about to write %d bytes\n", fd, (int)count );
+    _gpgme_debug (2, "fd %d: write `%.*s'\n", fd, (int) count, buffer);
     if ( !c ) {
         DEBUG0 ( "no writer thread\n");
         return -1;
@@ -646,9 +659,9 @@ _gpgme_io_pipe ( int filedes[2], int inherit_idx )
     sec_attr.nLength = sizeof sec_attr;
     sec_attr.bInheritHandle = FALSE;
     
-    if (!CreatePipe ( &r, &w, &sec_attr, 0))
+    if (!CreatePipe ( &r, &w, &sec_attr, PIPEBUF_SIZE))
         return -1;
-    /* make one end inheritable */
+    /* Make one end inheritable. */
     if ( inherit_idx == 0 ) {
         HANDLE h;
         if (!DuplicateHandle( GetCurrentProcess(), r,
@@ -762,27 +775,36 @@ _gpgme_io_set_nonblocking ( int fd )
 static char *
 build_commandline ( char **argv )
 {
-    int i, n = 0;
-    char *buf, *p;
-
-    /* FIXME: we have to quote some things because under Windows the 
-     * program parses the commandline and does some unquoting */
-    for (i=0; argv[i]; i++)
-        n += strlen (argv[i]) + 2 + 1; /* 2 extra bytes for possible quoting */
-    buf = p = malloc (n);
-    if ( !buf )
-        return NULL;
-    *buf = 0;
-    if ( argv[0] )
+  int i, n = 0;
+  char *buf, *p;
+  
+  /* FIXME: we have to quote some things because under Windows the
+   * program parses the commandline and does some unquoting.  For now
+   * we only do very basic quoting to the first argument because this
+   * one often contains a space (e.g. C:\\Program Files\GNU\GnuPG\gpg.exe) 
+   * and we would produce an invalid line in that case.  */
+  for (i=0; argv[i]; i++)
+    n += strlen (argv[i]) + 2 + 1; /* 2 extra bytes for possible quoting */
+  buf = p = malloc (n);
+  if ( !buf )
+    return NULL;
+  *buf = 0;
+  if ( argv[0] )
+    {
+      if (strpbrk (argv[0], " \t"))
+        p = stpcpy (stpcpy (stpcpy (p, "\""), argv[0]), "\"");
+      else
         p = stpcpy (p, argv[0]);
-    for (i = 1; argv[i]; i++) {
-        if (!*argv[i])
+      for (i = 1; argv[i]; i++)
+        {
+          if (!*argv[i])
             p = stpcpy (p, " \"\"");
-        else
+          else
             p = stpcpy (stpcpy (p, " "), argv[i]);
+        }
     }
-
-    return buf;
+  
+  return buf;
 }
 
 
@@ -891,13 +913,13 @@ _gpgme_io_spawn ( const char *path, char **argv,
         return -1;
     }
 
-    /* close the /dev/nul handle if used */
+    /* Close the /dev/nul handle if used. */
     if (hnul != INVALID_HANDLE_VALUE ) {
         if ( !CloseHandle ( hnul ) )
             DEBUG1 ("CloseHandle(hnul) failed: ec=%d\n", (int)GetLastError());
     }
 
-    /* Close the other ends of the pipes */
+    /* Close the other ends of the pipes. */
     for (i = 0; fd_parent_list[i].fd != -1; i++)
       _gpgme_io_close (fd_parent_list[i].fd);
 
@@ -1064,7 +1086,7 @@ _gpgme_io_select ( struct io_select_fd_s *fds, size_t nfds, int nonblock )
          */
         any = 0;
         for (i=code - WAIT_OBJECT_0; i < nwait; i++ ) {
-            if (WaitForSingleObject ( waitbuf[i], NULL ) == WAIT_OBJECT_0) {
+            if (WaitForSingleObject (waitbuf[i], 0) == WAIT_OBJECT_0) {
                 assert (waitidx[i] >=0 && waitidx[i] < nfds);
                 fds[waitidx[i]].signaled = 1;
                 any = 1;
@@ -1085,7 +1107,7 @@ _gpgme_io_select ( struct io_select_fd_s *fds, size_t nfds, int nonblock )
             int k, j = handle_to_fd (waitbuf[i]);
                     
             DEBUG1 ("WFMO invalid handle %d removed\n", j);
-            for (k=0 ; k < nfds; i++ ) {
+            for (k=0 ; k < nfds; k++ ) {
                 if ( fds[k].fd == j ) {
                     fds[k].for_read = fds[k].for_write = 0;
                     goto restart;
@@ -1116,3 +1138,11 @@ _gpgme_io_select ( struct io_select_fd_s *fds, size_t nfds, int nonblock )
     
     return count;
 }
+
+void
+_gpgme_io_subsystem_init (void)
+{
+  
+}
+
+
